############################################################################
# 
#  Insidious Mail DB
#
#-------------------
# Copyright 1996, Xerox Corporation.  All Rights Reserved.
# License is granted to copy, to use, and to make and to use derivative works for
# research and evaluation purposes, provided that the copyright notice and
# this license notice is included in all copies and any derivatives works and in
# all  related documentation.  Xerox grants no other licenses expressed or
# implied and the licensee acknowleges that Xerox have no liability for
# licensee's use or for any derivative works made by licensee. The Xerox
# names shall not be used in any advertising or the like without their written
# permission.
# This software is provided AS IS.
# XEROX CORPORATION DISCLAIMS AND LICENSEE
# AGREES THAT ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
# NOTWITHSTANDING ANY OTHER PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES
# RESULTING FROM THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, INCLUDING
# CONSEQUENTIAL OR ANY OTHER INDIRECT DAMAGES, WHETHER ARISING IN CONTRACT, TORT
# (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF XEROX CORPORATION
# IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
#---------
# This package saves the e-mail address of everyone you get mail from, and lets
# you send mail back with only a partial address.
#
# When you type part of an address in the To: or Cc: field, Ctrl-TAB will
# attempt to complete the address.
#
# The concept owes a lot to the Gnu Emacs package "BBDB" by Jamie Zawinski 
# (jwz@netscape.com) but this implementation is strictly my own.  Thanks, Jamie!
#
# To install, put this in your exmh-hacks directory (as shown in the 
# preferences menu) and put an invocation of Addr_Init in your User_Init
# proc (or use the one in the distributed user.tcl).  Do a MakeTclIndex and 
# restart exmh (or use tkinspect to send "Addr_Init" to a running exmh.
#
#
# If you put this:
#    *Sedit.Menubar.ubuttonlist: addrbrowse
#    *Sedit.Menubar.addrbrowse.text: Browse Addrs
#    *Sedit.Menubar.addrbrowse.command: Addr_Browse
#    *Mops.ubuttonlist: addrbrowse 
#
# in your .exmh-defaults file, 
# you will get a button on your message button line labled "Addrs"; this will 
# let you browse the address database.  One neat feature 
#
#  If you turn Addr_debug on, you will get some printing to exmh's stdout
# whenever a completion is attempted.  In addition, when you call up the Addrs
# browser you will get a LdSrc button that wil reload the addr.tcl file, if you decide
# hacking on this is fun.  Alas, the path to my copy is hard coded in the routine; 
# this should be fixed.
#
# A neat feature of the browser (and entry editor) is that you can pop up (a la Clip)
# the last message you got from that person.  Be careful, though; if that message was 
# deleted or the folder was packed since that message arrived, it will fail.
#
# I've been using this (or earlier versions) for about 2 years now and my database 
# is about 5000 entries; I find it EXTERMELY useful when I cannot quite remember the email
# address of that guy I got a message from 6 moonths ago and need to reply to, but
# I remember it was "Ted something".
#
# It takes a little while to load the browser window (at least for me, sorting 5000 
# strings and then inserting them into a listbox takes a while), but then it "stays"
# even if you dismiss it so it's not too painful.
#
# Enjoy; if you find this useful please let me know; iif you make it better 
# please send me the code.
#
#   --Berry Kercheval, Xerox PARC, March 1996 (kerch@parc.xerox.com)
#########################################################################

#
# These variables control the setup
#
set addrVersion {$Revision: 1.6 $}

# Addr_debug, if ==1, enables printing messages while this stuff runs.
# Set Addr_debug 1 in your user.tcl before invoking Addr_Init to enable debug messages
if {0 == [info exists Addr_debug]} {
    set Addr_debug 0
}

#
# Addr_Init loads the database file at startup time, and arranges the 
# partial-address-expansion keybinding.
#
proc Addr_Init {} {
    global env
    global addrFile
    global homeDir
    global addr_list

    # addrFile is the name of the file the address database is kept in
    set addrFile ".exmh_addrs"
    
    # homeDir is the directory the database is kept in.
    set homeDir $env(HOME)

    # tell exmh about our preference items and initialize them
    Preferences_Add "Address Database" "These settings affect the behavior of the address database.
See also the key binding for \"addrexpand\" that is set
in the Bindings... dialog Simple Edit." {
        {
            addr_db(enabled)
            addressdbEnabled
            ON
            "Enable address saving"
            "If set, From addresses are remembered and available in an address browser."
        }
        {
            addr_db(hideexcluded)
            addressdbHideExcluded
            ON
            "Hide excluded addresses"
            "If set, addresses excluded from the expansion process will initially not be displayed by the Address DB Browser."
        }
        {
            addr_db(checkpoint_on_folder_change) 
            addressdbFolderChangeCheckpoint
            ON
            "Checkpoint on Folder Change"
            "If set, Exmh will save your address file whenever you visit a new folder."
        }
        {
            addr_db(searchlist)
            addressdbSearchlist
            {Addr_FullNameMatch Addr_Lookup Alias_Lookup}
            "Expand methods to use"
            "A list of TCL procs, separated by spaces, which will be called sequentially to try to expand the address.  Valid choices are \"Addr_FullNameMatch\" to search full names, \"Addr_Lookup\" to search for mail addresses, and \"Alias_Lookup\" to search your MH/exmh alias list."
        }
        {
            addr_db(filter_regexp)
            addressdbFilterRegexp
            {}
            "Regular expression filter"
            "If set, addresses matching this regular expression pattern will not be saved in the database."
        }
        {
            addr_db(skip_folders)
            addressdbFoldersSkip
            {}
            "Folders to ignore"
            "A list of one or more folders separated by spaces.  Exmh will not save addresses from the mail in the folders in this list. An empty list will allow Exmh to add addresses from every folder.  Groups of folders may be specified using * as a wild card anywhere in a folder name."
        }
        {
            addr_db(filter_alternate_mailboxes) 
            addressdbFilterAltMailboxes
            ON
            "Ignore alternate mailboxes"
            "If set, addresses that match the names specified as \"Alternate mailboxes\" in your MH profile will not be saved."
        }
        {
            addr_db(key_force_save)
            addressdbForceSave 
            <Control-Tab>
            "Key to save an address"
            "Key which, if pressed, will cause the address from the current message to be stored regardless of any filtering specified.  This key is only active in the main exmh window so it may be the same as the \"Key to expand addresses\" without conflict.  Pressing this key stores an address in the database even if it would have been filtered (not stored) due to matching \"Regular expression filter\", \"Folders to ignore\", or one of your alternate mailboxes."
        }
    }

    #addr_db is an array used for keeping state.
    #make this an array from the get-go and set the default pref...
    global addr_db
    set addr_db(init) 1;  
    set addr_db(curmethod) 0
    set addr_db(laststring) ""
    set addr_db(lastfound) ""
    set addr_db(changed) 0
    Addr_LoadDB
}
#
# Hook_MsgShow is called when exmh displays a message.  We parse out the 
# From: header and call Addr_Save to update the entry in the database.
#
proc Hook_MsgShowAddr {path headervar } {
    upvar $headervar header
    global addr_db

    set addr_db(last_seen) [list $path $header(0=1,hdr,from) $header(0=1,hdr,date)]
    if {! $addr_db(enabled)} {
	return
    }
    Addr_Save [MsgParseFrom $header(0=1,hdr,from)] $path \
            $header(0=1,hdr,from) 	$header(0=1,hdr,date)
}

#
# Hook_FolderChange is called when a new folder is visited.  Save the 
# database then too.
#
proc Hook_FolderChangeAddr {newfolder} {
    global addr_db

    if {$addr_db(changed) && $addr_db(checkpoint_on_folder_change)} {
		Addr_SaveFile
    }
}


#
# Hook_CheckPoint is called when exmh checkpoints its state.  Save the 
# database here.
#
proc Addr_CheckPoint {} {
	Addr_SaveFile
}

#####
#
# Address_Init and the address Hook procedures are moved to extrasInit.tcl
#
####

########################################################################
#
# This one is bound to a key of the user's choosing to force-save 
# the from address of the current message (info stored in last-seen by
# the Hook_MsgShow routine above.
#
proc Address_Save {} {
    global addr_db

    AddrDebug "Force save $addr_db(last_seen)"

    Addr_Save [MsgParseFrom [lindex $addr_db(last_seen) 1]] \
            [lindex $addr_db(last_seen) 0] \
            [lindex $addr_db(last_seen) 1] \
            [lindex $addr_db(last_seen) 2] \
            "force"
}

########################################################################
#
# These are the "real" database procs.
#

#
# This is "magically" executed when exmh is setting up 
# keystroke bindings for the main window
#
proc User_Bindings { w } {
    global addr_db
    if {$addr_db(key_force_save) != ""} {
        AddrDebug "binding for $w"
        Bind_Key $w $addr_db(key_force_save) {Address_Save ; break}
    }
}

#
# Addr_LoadDB does the real work to load the database file.
#
proc Addr_LoadDB {} {
    global addr_db
    global addrFile
    global homeDir
    global addr_list

    AddrDebug "AddrDB: loading database..."
    catch {source $homeDir/$addrFile}
    set addr_db(changed) 0
    AddrDebug "AddrDB: loading database...done."
    # need to update listbox...
}
#
# Addr_SaveFile saves the database into a unix file 
#
proc Addr_SaveFile {} {
    global addr_list addr_db
    global addrFile
    global homeDir

    Exmh_Status "Saving address database..."
    set fd [open "$homeDir/.exmh_addr_tmp" w]
    foreach i [array names addr_list] {
        puts $fd "set addr_list($i) {$addr_list($i)}"
    }
    close $fd
    #  the first time the address file won`t exist yet...
    if [file exists $homeDir/$addrFile] {
        Mh_Rename $homeDir/$addrFile $homeDir/$addrFile.bak
    }
    Mh_Rename $homeDir/.exmh_addr_tmp $homeDir/$addrFile 
    set addr_db(changed) 0
    Exmh_Status "Saving address database...done."
}

#
# Addr_Save updates the database entry when a new message is read.
#

proc Addr_Save {from path rawfrom date {forcesave "not"}} {
    global addr_db
    global addr_list
    global exmh
    global mhProfile

    if {[string compare $forcesave "force"] != 0} {
        if {$addr_db(skip_folders) != ""} {
            set filter_list [split $addr_db(skip_folders) " "]
            # AddrDebug "  folder filter list is \"$filter_list\""
            foreach i $filter_list {
                if {[string length $i] > 0} {
                    # AddrDebug "Matching \"$exmh(folder)\" for \"$i\""
                    if [string match $i $exmh(folder)] {
                        AddrDebug "  folder filter eliminated $exmh(folder)"
                        return
                    }
                }
            }
        }
        if {$addr_db(filter_alternate_mailboxes) != 0} {
	    catch {unset filter_list}
            catch {
                set filter_list [split $mhProfile(alternate-mailboxes) ", "]
            }
            if [info exists filter_list] {
                # Filter out all of the user's alternate mailboxes
                # AddrDebug "alternate mailboxes \"$mhProfile(alternate-mailboxes)"
                set filter_list [split $mhProfile(alternate-mailboxes) ", "]
                # AddrDebug "  list \"$filter_list\""
                foreach i $filter_list {
                    if {[string length $i] > 0} {
                        if [string match $i $from] {
                            AddrDebug "  alternate mailbox filter eliminated $from"
                            return
                        }
                    }
                }
            }
        }
        if {$addr_db(filter_regexp) != ""} {
            # AddrDebug "filtering $addr_db(filter_regexp)"
            if [regexp -nocase $addr_db(filter_regexp) $from] {
                AddrDebug "  regexp filter eliminated $from"
                return
            }
        }
    }
    set addr_db(changed) 1
    set newone [list $path $date  \
			      [Addr_ParseFrom $rawfrom] $rawfrom] 
    if [info exists addr_list($from)] {
        Exmh_Debug "Updating address \"$from\"."
    } else {
        Exmh_Debug "Saving address \"$from\"."
        catch {
            set fullname  [lindex $newone 2]
            regsub ,$ $fullname {} fullname_less_comma
            $addr_db(win) insert end [format "%-25.25s %s" $fullname_less_comma $from ]
        }
    }
    set addr_list($from) $newone
}

#
# Addr_Expand expands partial addresses.
#
proc Addr_Expand {} {
    AddrDebug "Expanding address..."
}

#
# Addr_KeyExpand expands a partial address in response to a key binding
#
proc Addr_KeyExpand { w } {
    global addr_db

    AddrDebug "Addr_KeyExpand: w=$w $addr_db(searchlist)"
    set line [string trim [$w get {insert linestart} {insert lineend}]]
    ##  AddrDebug "  got line \"$line\""
    # Only allows expansion on addressable header lines.
    if [regexp -nocase {^(to: *|resent-to: *|cc: *|resent-cc: *|bcc: *|dcc: *)(.*)} $line t0 t1 t2] {
        ##  AddrDebug  "  matched! keep is \"$t1\", partial name=\"$t2\""
	if [regexp -indices ",?.*, *" $t2 t0] {
	    set t0 [lindex $t0 end]
            ##  AddrDebug "got comma at $t0"
            set t3 [string range $t2 0 $t0]
            append t1 $t3
            set t2 [string range $t2 [expr $t0 + 1] end]
            ##  AddrDebug "  multi, will keep \"$t1\", new partial name=\"$t2\""
        }
        if {[string compare $addr_db(lastfound) $t2] != 0 \
                || $addr_db(curmethod) >= [llength $addr_db(searchlist)]} {
            Exmh_Status "Reseting start method"
            catch {destroy $w.addrs}
            set addr_db(expansion) {}
            set addr_db(curmethod) 0
            set addr_db(laststring) $t2
        } else {
            if {$addr_db(curmethod) != 0} {
            	set t2 $addr_db(laststring)
            }
        }
        foreach proc [lrange $addr_db(searchlist) $addr_db(curmethod) end] {
            incr addr_db(curmethod)
            Exmh_Status "$proc $t2"
			set result [busy $proc $t2]
			if {[string compare $result ""] == 0} continue
            if {[llength $result] == 1} {
                # unique match
                $w delete  {insert linestart} {insert lineend}
                $w insert insert [format "%s%s" $t1 [lindex $result 0]]
                set addr_db(lastfound) [lindex $result 0]
                catch {destroy $w.addrs}
            } else {
                # must be multiple hits
                AddrDebug "  Multiple hits: $result"
                set addr_db(lastfound) $t2
                set new [AddrShowDialog $w $result]
                # if no selection is made, leave the string where it is
                if [ string compare $new "" ] {
                    set addr_db(lastfound) $new
                    $w delete  {insert linestart} {insert lineend}
                    $w insert insert [format "%s%s" $t1 $new]
                }
            }
			break
        }
    } else {
        Exmh_Status "Error in name expansion: not on To: field"
        return
    }
}

proc Alias_Lookup {n} {
    global aliases
	Aliases_Load
	set t2 [string trim $n]
	if {[string length [array names aliases $t2]] == 0} {
		Exmh_Status "No match found for \"$t2\""
		puts stdout "\a" nonewline
		return {}
	} {
		Exmh_Status "Found alias: \"$aliases($t2)\""
		return $aliases($t2)
	}
}

proc Addr_FullNameMatch {name }  {
    global addr_list

    Exmh_Status "Matching on full names with $name..."
    set result {}
    set pat {}
    append pat $name
    foreach i [ array names addr_list] {
        set elt $addr_list($i)
        if {[lindex $elt 4] == 1} continue
        set fn [lindex $elt 2]
        # puts stdout "matching against $fn (elt = $elt)"
        if [regexp -nocase $pat $fn t0] {
            AddrDebug "   fullname match on $i ($fn)"
            lappend result "$i ($fn)"
        }  
    }
    if {[llength $result] > 0} {
        return $result
    } else {
        Exmh_Status "Matching on full names with $name...none found"
        return {}
    }
}

proc Addr_Lookup { n } {
    global addr_list

    AddrDebug "Addr_Lookup: looking for $n"
    if {[string compare $n ""] == 0} {
        Exmh_Status "Address expansion error: null string!"
        return {}
    }
    set result {}

    set pat {}
    append pat "^" $n
    AddrDebug "  using pattern \"$pat\""

    foreach i [array names addr_list] {
        set elt $addr_list($i)
        if {[lindex $elt 4] == 1} continue
        if [regexp -nocase $pat $i t0] {
            AddrDebug "   match on $i"
            set fn [lindex $elt 2]
            if {[string compare $fn ""]} {
                lappend result "$i ($fn)"
            } else {
                lappend result "$i"
            }
        }
    }

    AddrDebug "Addr_Lookup: result is $result"
    return $result
}
#
# Addr_ParseFrom takes a raw From: header and return the fullname.
# it should work on lines of the form:
#
## Berry Kercheval <kerch@parc.xerox.com>
## kerch@parc.xerox.com (Berry Kercheval)

proc Addr_ParseFrom { fromline } {
    #    AddrDebug "Addr_ParseFrom: working on $fromline"
    set line [string trim $fromline]

    # if it's "xxx <foo@bar>"...
    if [regexp {([^<]*)(<.*>)} $line t1 t2 t3] {
        #	AddrDebug "  Matched: ( $t1 )( $t2 )( $t3 )"
        set token [string trim $t2 ]
    } else {
        # nope, try foo@bar (xxx)
        #	AddrDebug "  Not xxx <foo@bar>, try foo@bar (xxx)"
        if [regexp {^([^\(]*)(\(.*\))[^\)]*$} $line t1 t2 t3] {
            #	    AddrDebug "  Matched: ( $t1 )( $t2 )( $t3 )"
            set token $t3
        } else {
            # none of the above, give up.
            set token {}
        }
    }
    
    #    AddrDebug "  result is $token"
    set token [string trim $token "\"()"]
    #    AddrDebug "  trimmed result is $token"
    return $token
}

#
# Debug support
#

proc AddrDebug { s {nonewline {}}} {
    global Addr_debug
    if {$Addr_debug == 1} {
        if {[string compare $nonewline ""] == 0} {
            puts stdout $s 
        } else {
            puts stdout $s nonewline
        }
    }
}

proc AddrShowDialog {w list} {
    global addr_db

    catch {destroy $w.addrs}
    set f [frame $w.addrs -bd 4 -relief ridge]
    set l [listbox $f.lb -bd 4 -width 50 -height 10]
    bind $l <Any-Double-1> "\
      AddrShowDialogDone $f $l ;\
      break \
    "

    foreach i $list {
        $l insert end $i
    }
    pack $f.lb -expand true -fill both
    frame $f.but -bd 10 -relief flat
    pack $f.but -expand true -fill both
    Widget_AddBut $f.but ok "Done" [list AddrShowDialogDone $f $l] left
    Widget_PlaceDialog $w $f
    tkwait window $f
    if [info exists addr_db(expansion)] {
    	Exmh_Status "returning $addr_db(expansion)"
    	return $addr_db(expansion)
    } else {
		return {}
    }
}

proc AddrShowDialogDone {f l} {
    global addr_db
    set result [$l curselection]
    if {[string compare $result ""] != 0} {
	set result [lindex $result 0]
        set name [$l get $result]
        AddrDebug "Selected: $result ($name)"
        set addr_db(expansion) $name
    } else {
        AddrDebug "Selected: <nothing>"
        set addr_db(expansion) ""
    }
    AddrDebug "selected $addr_db(expansion)"
    catch {destroy $f}
}

proc Addr_Browse {} {
    global addr_br 
    global addr_db 
    global addr_list
    global Addr_debug

    set t .addr_br
    set f .addr_br.but
    if [Exwin_Toplevel .addr_br "Address DB Browser" Addr_Br] {
        Exmh_Status "Creating Address Browser..."

        .addr_br.but.quit configure -command {Exwin_Dismiss .addr_br}

        Widget_AddBut $f reload "ReLoad" { Addr_Browse_Reload }
        Widget_AddBut $f save   "Save"   { Addr_Browse_Save }
        Widget_AddBut $f delete "Delete" { Addr_Browse_Delete }
        Widget_AddBut $f edit   "Edit"   { Addr_Browse_Edit }
        Widget_AddBut $f view   "ViewLastMsg" { Addr_Browse_Clip }
        if {$addr_db(hideexcluded) == 0} {
		Widget_AddBut $f hide   "Hide Excluded" { Addr_Browse_Toggle }
        } else {
		Widget_AddBut $f hide   "Show Excluded" { Addr_Browse_Toggle }
        }

        if { $Addr_debug == 1 }  {	Widget_AddBut $f ldsrc  "LdSrc"  { Addr_Load_Source } }

        # should have menu to sort by fullname, date, addr, folder, etc
        set addr_db(win) [listbox $t.lb -height 20 -width 65 \
                -relief sunken \
                -yscrollcommand [list $t.sb set] ]
        scrollbar $t.sb -orient vertical -command [list $addr_db(win) yview] 
        
        bind $addr_db(win) <Any-Double-1>    {Addr_Browse_Edit}
        bind $addr_db(win) <B1-Motion>       {%W select set [%W nearest %y]}
        ##  bind $addr_db(win) <Shift-B1-Motion> {%W select from [%W nearest %y]}
        bind $addr_db(win) <Shift-Button-1>  {%W select set [%W nearest %y]}
        bind $addr_db(win) d                 {Addr_Browse_Delete}
        pack $t.sb -side left -fill y
        pack $addr_db(win) -expand true -fill both
    } else {
	$addr_db(win) delete 0 end
    }
    Exmh_Status "Creating Address Browser...getting names..."
    update idletasks

    if {[array size addr_list] == 0} {
	set addr_list(noone@nowhere.nada.zip) 1; #null array, put in a dummy
    }
    foreach i [array names addr_list] {
	set fullname  [lindex $addr_list($i) 2]
	regsub ,$ $fullname {} fullname_less_comma
	if {[lindex $addr_list($i) 4] == 1} {
	    if {$addr_db(hideexcluded) != 1} {
		    lappend l [format "%-24.24s- <%s>" $fullname_less_comma $i ]
	    }
	} else {
	    lappend l [format "%-25.25s <%s>" $fullname_less_comma $i ]
	}
	#lappend l [format "%-35s %s" $i $fullname_less_comma ]
    }
    Exmh_Status "Creating Address Browser...sorting names..."
    update idletasks
    set l [lsort $l]
    set msg "Creating Address Browser...inserting names..."
    set n 0
    set whiz [list | \\ - /]
    set w 0
    foreach i $l {
	$addr_db(win) insert end $i
	incr n
	if { 0==($n%100) } {
	    Exmh_Status [concat $msg [lindex $whiz $w]]
	    update idletasks
	    set w [expr {($w+1)%4}]
	}
    }    
    Exmh_Status "Creating Address Browser...done"
}

proc Addr_Browse_Toggle {} {
    global addr_db
	if {$addr_db(hideexcluded) == 0} {
		set addr_db(hideexcluded) 1
	} else {
		set addr_db(hideexcluded) 0
	}
	destroy .addr_br
	Addr_Browse
}

proc Addr_Browse_Clip {} {
	global addr_db addr_list mhProfile

	set sel [lindex [$addr_db(win) curselection] 0]
	# Check that selection is not null....
	if { $sel == {} } {
		return
	}

    set victim [string trim [string range [$addr_db(win) get  $sel] 25 end] { <>}]
	set xx $addr_list($victim)
	set last [lindex $xx 0]	

	# since there may be recursive folders, match the MH Path
	# off the front, the message number off the end and the rest must be the folder.

	set pat "($mhProfile(path))/(.+)/(\[0-9\]+)\$"
	if [regexp $pat $last match path folder msg] {
	    Msg_Clip $folder $msg
	} else {
	    Exmh_Status "Cannot find $last"
	}
}

proc Addr_Browse_Reload {} {
    global addr_db addr_list

    if $addr_db(changed) {
        if [Addr_Browse_ChangedDialog $addr_db(win)] {
            Exmh_Status "Reloading address database..."
            unset addr_list
            set addr_list(noone@nowhere.nada.zip) 1; #make this an array too.
            Addr_LoadDB
            Exmh_Status "Reloading address database...done."
        }  else {
            Exmh_Status "Reload of address database aborted."
        }
    } else { 
        Exmh_Status "Reloading address database..."
        unset addr_list
        set addr_list(noone@nowhere.nada.zip) 1; #make this an array too.
        Addr_LoadDB
        Exmh_Status "Reloading address database...done."
    }
}

proc Addr_Browse_ChangedDialog {w} {
    global addr_db
    set f [frame $w.addrch -bd 4 -relief ridge]
    Exmh_Status "$f"
    Widget_AddBut $f ok "Yes, this will lose all changes" "\
        set addr_db(changeresult) 1 ;\
        destroy $f " \
        left
    Widget_AddBut $f no "No, do not reload" "\
        set addr_db(changeresult) 0 ;\
        destroy $f " \
        right
    Widget_PlaceDialog $w $f
    tkwait window $f
    AddrDebug "Addr_Browse_ChangedDialog returns  $addr_db(changeresult)"
    return $addr_db(changeresult)
}

proc Addr_Browse_Save {} {
    
    Addr_SaveFile
}

proc Addr_Browse_Delete {} { 
    global addr_db addr_list

    foreach sel [lsort -decreasing [$addr_db(win) curselection]] {
	set victim [string trim [string range [$addr_db(win) get  $sel] 25 end] { <>}]
	if [catch {unset addr_list($victim)} val] {
	    Exmh_Status "Address DB: can't delete: $val"
	} else {
	    $addr_db(win) delete $sel
	    Exmh_Status "Address DB: Deleted $victim"
	}
    }
}

proc CommandEntry { name label width command  } { 
    frame $name
    label $name.label -text $label -width $width -anchor w
    eval {entry $name.entry -relief sunken -width  50 } 
    pack $name.label -side left
    pack $name.entry -side right -fill x -expand true
    bind $name.entry <Return> $command
    return $name.entry
}

proc Addr_Browse_Edit {} {
    global addr_db addr_list

    set  sel [lindex [$addr_db(win) curselection] 0]
    # Check that selection is not null....
    if { $sel == {} } {
        return
    }

    set victim [string trim [string range [$addr_db(win) get  $sel] 25 end] { <>}]
    set xx $addr_list($victim)
    set name [lindex $xx 2]
    set addr [lindex $xx 3]
    set last [lindex $xx 0]
    set date [lindex $xx 1]
    set exclude [lindex $xx 4]
	Exmh_Status "Exclude $exclude"
    if [info exists .addr_ed] {
        puts stdout "got a window already"
    }
    if [Exwin_Toplevel .addr_ed "Address DB editor" Addr_Ed] {
        Exmh_Status "Creating Address Editor for $victim"
        set t .addr_ed
        set f .addr_ed.but

        .addr_ed.but.quit configure -command { Addr_Edit_Dismiss }
        $t configure -width 500

        Widget_AddBut $f save   "Save"   { Addr_Edit_Save }
        Widget_AddBut $f delete "Delete" { Addr_Edit_Delete }	
        Widget_AddBut $f last	"ViewLastMsg"	 { Addr_Browse_Clip }
        set e [Widget_AddBut $f ignore	"Exclude"	 { Addr_Exclude }]

        set n [CommandEntry $t.name		"Full Name"		12  Addr_Edit_Save ]
        set a [CommandEntry $t.address	"Address"		12 	Addr_Edit_Save ]
        set l [CommandEntry $t.lastMsg	"Last Message"	12 	Addr_Edit_Save ]
        set d [CommandEntry $t.date		"Date"			12 	Addr_Edit_Save ]
        
        pack $t.name $t.address	 $t.lastMsg  $t.date

    } else {

        set wins $addr_db(editwin)
        set n [lindex $wins 0]
        set a [lindex $wins 1]
        set l [lindex $wins 2]
        set d [lindex $wins 3]
        set e [lindex $wins 5]

    }

    set addr_db(editwin) [list $n $a $l $d $sel $e]

    $n delete 0 end;    $n insert 0 $name
    $a delete 0 end;    $a insert 0 $addr
    $l delete 0 end;    $l insert 0 $last
    $d delete 0 end;    $d insert 0 $date
	
    if {$exclude == 1} {
       	$e config -text "Include"
        Exmh_Status "$e config -text Include"
    } else {
       	$e config -text "Exclude"
        Exmh_Status "$e config -text Exclude"
    }
}


proc Addr_Load_Source {}  {
    global env
    # HACK HACK HACK!!!
    source "~/.tk/exmh/addr.tcl"
}

########################################################################
#
# Address Editor Routines
#

proc Addr_Edit_Clip {s} {
    global mhProfile

    # since there may be recursive folders, match the MH Path
    # off the front, the message number off the end and the rest must be the folder.

    set pat "($mhProfile(path))/(\[a-zA-Z_/-\]+)/(\[0-9\]+)\$"
    regexp $pat $s match path folder msg
    Msg_Clip $folder $msg
}

proc Addr_Exclude {} {
    global addr_db addr_list

    # entries are: name address lastmsg date
    set wins $addr_db(editwin)
    set name [[lindex $wins 0] get]
    set addr [[lindex $wins 1] get]
    set last [[lindex $wins 2] get]
    set date [[lindex $wins 3] get]
    set sel [lindex $wins 4]

    set index [MsgParseFrom $addr]
    if {[lindex $addr_list($index) 4] == 1} {
		set exclude 0
    } else {
		set exclude 1
    }
    # NEED TO STUFF new NAME into entry!
    set addr [format "%s <%s>" $name $index]
	Exmh_Status "Updating address \"$index\"."
    set addr_db(changed) 1
    set addr_list($index) [list $last $date  \
            [Addr_ParseFrom $addr] $addr $exclude] 

    # update browser window...

    $addr_db(win) delete $sel

    if {$addr_db(hideexcluded) == 0 || $exclude == 0} {
        set entry $addr_list($index)

    	set fullname  [lindex $entry 2]
    	regsub ,$ $fullname {} fullname_less_comma
	 	if {[lindex $entry 4] == 1} {
            set new [format "%-24.24s- <%s>" $fullname_less_comma $index ]
	 	} else {
            set new [format "%-25.25s <%s>" $fullname_less_comma $index ]
	 	}
    	#set new [format "%-35s %s" $index $fullname_less_comma ]
        
    	$addr_db(win) insert $sel $new
    }
    # make it all go away so we can redo it next time.
    Addr_Edit_Dismiss
}

proc Addr_Edit_Save {} {
    global addr_db addr_list

    # entries are: name address lastmsg date
    set wins $addr_db(editwin)
    set name [[lindex $wins 0] get]
    set addr [[lindex $wins 1] get]
    set last [[lindex $wins 2] get]
    set date [[lindex $wins 3] get]
    set sel [lindex $wins 4]

    set index [MsgParseFrom $addr]
    # NEED TO STUFF new NAME into entry!
    set addr [format "%s <%s>" $name $index]
	Exmh_Status "Updating address \"$index\"."
    set addr_db(changed) 1
    set addr_list($index) [list $last $date  \
            [Addr_ParseFrom $addr] $addr] 

    # update browser window...

    $addr_db(win) delete $sel

    set entry $addr_list($index)

    set fullname  [lindex $entry 2]
    regsub ,$ $fullname {} fullname_less_comma
    set new [format "%-25.25s <%s>" $fullname_less_comma $index ]

    $addr_db(win) insert $sel $new

    # make it all go away so we can redo it next time.
    Addr_Edit_Dismiss
}

proc Addr_Edit_Delete {} {
    global addr_db addr_list

    set wins $addr_db(editwin)
    set addr [[lindex $wins 1] get]
    set sel [lindex $wins 4]
    set index [MsgParseFrom $addr]
    $addr_db(win) delete $sel

    unset addr_list($index)
    Addr_Edit_Dismiss
}

proc Addr_Edit_Abort {} {
    global addr_db
    Addr_Edit_Dismiss
}

proc Addr_Edit_Dismiss {} {
    Exwin_Dismiss .addr_ed
}
########################################################################
#
# done loading...
#
if {$Addr_debug} { puts stdout "done." }

